; ==============================================
; pexl RTL (Windows x86 version)
; DEVELOPMENT COPY ONLY
; ==============================================
; Platform-specific notes:
;
; pxobject header is 12 bytes stored immediately preceding the returned pointer
;  containing only the reference count

; pxstring := (refct allocation_size destructor_fun) strbuff strbuffsize strlen
;  Using double indirection to allow for resizing without impacting the actual pointer
;  24 bytes of overhead for every string? C wins :)

; Automatic and deterministic destructor calls (like Python, unlike D or Java)

	extern _HeapCreate@12
	extern _HeapAlloc@12
	extern _HeapReAlloc@16
	extern _HeapFree@12

	; For pxo_string_trace
	extern _GetStdHandle@4
	extern _WriteFile@20
	
;
; Macros
;
	
%define PXO_SIZE              12
%define PXO_OFFSET_REFCT      0
%define PXO_OFFSET_ALLOCSIZE  4
%define PXO_OFFSET_DESTRUCTOR 8

%define PXSTRING_SIZE 12
%define PXSTRING_OFFSET_BUFFER   0
%define PXSTRING_OFFSET_BUFFSIZE 4
%define PXSTRING_OFFSET_STRLEN   8

%define PUBLIC_OBJECT \
	dd 0, 0, 0
	
; pxstr	
%macro pxstring 2
	PUBLIC_OBJECT
%1:
	%strlen last_strlen %2
	dd $ + PXSTRING_SIZE
	dd last_strlen ; buffsize
	dd last_strlen ; strlen
	dd %2
	%undef last_strlen
%endmacro
	
;
; Consts
;
	
	section .data
	
pxl_rtl_heap:  dd 0
pxl_rtl_xheap: dd 0	

	pxstring hello_pexl, "PEXL"

;
; RTL Core
;

	section .text use32

; void rtl-init()
; Allocate rw and rwx heaps for malloc/xalloc
pxl_rtl_init:
	push ebp
	mov ebp, esp
	
	push 0 ; dwMaximumSize = NULL
	push 0 ; dwInitialSize = NULL (one page)
	push 0 ; flags = 0
	call _HeapCreate@12
	mov dword [pxl_rtl_heap], eax
	
	push 0 ; dwMaximumSize = NULL
	push 0 ; dwInitialSize = NULL (one page)
	push 0x00040000 ; flags = HEAP_CREATE_ENABLE_EXECUTE
	call _HeapCreate@12
	mov dword [pxl_rtl_xheap], eax
		
	leave ; mov esp, ebp && pop ebp
	ret 0 ; Done
	
; void* __stdcall xalloc(size_t len)
; @param   len  Amount of bytes to allocate
; @return       Pointer to executable memory at least $len bytes, or 0 on failure
pxl_rtl_xalloc:
	push ebp
	mov ebp, esp
	
	push dword [ebp+8]           ; $len (SIZE_T dwBytes)
	push 0                       ; DWORD dwFlags
	push dword [pxl_rtl_xheap]   ; HANDLE hHeap
	call _HeapAlloc@12

	; Return is in eax, leave it there
	leave
	ret 4                 ; one argument

; void __stdcall pxl_rtl_xfree(void* ptr)
; @param   ptr  Pointer to memory to free
pxl_rtl_xfree:
	push ebp
	mov ebp, esp

	push dword [ebp+8]
	push 0
	push dword [pxl_rtl_xheap]
	call _HeapFree@12
	
	leave
	ret 4                 ; one argument
	
; void* __stdcall malloc(size_t len)
; @param   len  Amount of bytes to allocate
; @return       Pointer to executable memory at least $len bytes, or 0 on failure
pxl_rtl_malloc:
	push ebp
	mov ebp, esp
	
	push dword [ebp+8]           ; $len (SIZE_T dwBytes)
	push 0                       ; DWORD dwFlags
	push dword [pxl_rtl_heap]    ; HANDLE hHeap
	call _HeapAlloc@12

	; Return is in eax, leave it there
	leave
	ret 4                 ; one argument

; void __stdcall pxl_rtl_realloc(void* ptr, size_t len)
;  Resizes a pointer and returns its new memory location.
; @param void* ptr
; @param len
pxl_rtl_realloc:
	push ebp
	mov ebp, esp
	
	push dword [ebp+12] ; len
	push dword [ebp+8]  ; ptr
	push 0              ; flags
	push dword [pxl_rtl_heap]
	call _HeapReAlloc@16
	
	leave
	ret 8 ; two arguments
	
; void __stdcall pxl_rtl_free(void* ptr)
; @param   ptr  Pointer to memory to free
pxl_rtl_free:
	push ebp
	mov ebp, esp

	push dword [ebp+8]
	push 0
	push dword [pxl_rtl_heap]
	call _HeapFree@12
	
	leave
	ret 4                 ; one argument
	
; heapfn __stdcall bind1st(void* func, void* param1)
; @param   func    Pointer to function to generate a stub for
; @param   param1  Parameter to bind into the stub
; @return          Executable stub with bound parameter, or 0 on failure
pxl_rtl_bind1st:
	push ebp
	mov ebp, esp
	
	; Allocate executable memory
	push 13
	call pxl_rtl_xalloc
	
	; Copy arguments to registers
	mov ebx, dword [ebp+8]
	mov ecx, dword [ebp+12]
	
	; Copy stub into memory, return is in eax
	push edx
	mov byte [eax], 91      ; 0x5B pop ebx
	mov byte [eax+1], 104   ; 0x68 XXXXXXXX push dword $param1
	mov edx, ecx
	mov byte [eax+2], dl
	mov edx, ecx
	sar edx, 8
	mov byte [eax+3], dl
	mov edx, ecx
	sar edx, 16
	mov byte [eax+4], dl
	mov edx, ecx
	shr edx, 24
	mov byte [eax+5], dl
	mov byte [eax+6], 83    ; 0x53 push ebx
	mov byte [eax+7], 104   ; 0x68 XXXXXXXX push dword $func
	mov byte [eax+8], bl
	mov edx, ebx
	sar edx, 8
	mov byte [eax+9], dl
	mov edx, ebx
	sar edx, 16
	mov byte [eax+10], dl
	shr ebx, 24
	mov byte [eax+11], bl
	mov byte [eax+12], -61  ; 0xC3 ret
	
	; Return was in eax, leave it there
	leave
	ret 8                   ; two arguments

; void memcpy(void* dest, void* source, uint numBytes)
; @param dest
; @param source
; @param numBytes
; @return void
pxl_rtl_memcpy:
	push ebp
	mov ebp, esp
	
	mov edi, dword [ebp+8]  ; dest
	mov esi, dword [ebp+12] ; src
	mov ecx, dword [ebp+16] ; count
	rep movsb	
	
	leave
	ret 12 ; three arguments
	
	
;
; PXObject
; 

	
; public void* pxo_new(size_t len)
;  Allocate a new PXObject
; @param  len   Number of octets to allocate
; @return void* Memory buffer (there's a pxobject header hidden 8 octets before)
PUBLIC_OBJECT
pxo_new:
	;
	push ebp
	mov ebp, esp
	push ebx	
	
	; eax := malloc( arg1 + PXO_SIZE )
	mov ebx, dword[ebp+8]
	add ebx, PXO_SIZE ; real alloced size
	push ebx
	call pxl_rtl_malloc
	
	; add default pxobj references
	mov dword[eax + PXO_OFFSET_REFCT], 1 ; refcount
	sub ebx, PXO_SIZE ; fake alloced size
	mov dword[eax + PXO_OFFSET_ALLOCSIZE], ebx ; buffsize
	mov dword[eax + PXO_OFFSET_DESTRUCTOR], 0;
	
	add eax, PXO_SIZE ; return alloc+PXO_SIZE
	
	pop ebx
	leave
	ret 4 ; one argument
	
; public void pxo_addref(pxobj*)
;  Increment reference count for a PXObject
; @param pxobj* A pointer to a buffer allocated with pxo_new
; @return void
PUBLIC_OBJECT
pxo_addref:
	;
	push ebp
	mov ebp, esp
	push ebx
	
	; ebx = &refct
	mov ebx, dword[ebp+8]
	sub ebx, PXO_SIZE - PXO_OFFSET_REFCT
	
	; if (*ebx != dword 0) {
	cmp dword[ebx], 0
	je pxo_addref_nocmp
	
	; refcount++
		inc dword[ebx]
	
	; }
pxo_addref_nocmp:
	
	; return 0
	mov eax, 0
	
	pop ebx
	leave
	ret 4 ; one argument
	
; public void pxo_release(pxobj*)
;  Decrement reference count for a PXObject, free if necessary
; @param pxobj* A pointer to a buffer allocated with pxo_new
; @return void
PUBLIC_OBJECT
pxo_release:
	;
	push ebp
	mov ebp, esp
	push ebx
	
	; ebx = &refct
	mov ebx, dword[ebp+8]
	sub ebx, PXO_SIZE - PXO_OFFSET_REFCT
	
	; if (*ebx != dword 0) {
	cmp dword[ebx], 0
	je pxo_release_nocmp
	
	; refcount--
		dec dword[ebx]
	
	; if (*ebx == dword 0) {
		cmp dword[ebx], 0
		jne pxo_release_nocmp
	
	; destructor
		push dword[ebp+8]
		add ebx, PXO_OFFSET_DESTRUCTOR - PXO_OFFSET_REFCT ; ebx = &destructor
		cmp dword[ebx], 0
		je pxo_destructor_called
		
		;this->destruct(this)
		push dword[ebp+8]
		call dword[ebx]
		;ret
pxo_destructor_called:
		sub ebx, PXO_OFFSET_DESTRUCTOR ; ebx = &base
		
	; free
		push ebx
		call pxl_rtl_free
	
	; }}
pxo_release_nocmp:
	
	; return 0
	mov eax, 0
	
	pop ebx
	leave
	ret 4 ; one argument
	
	
;
; PXString
;
	
	
; public void pxo_string_new()
;  Allocate a new string with a null string buffer
; @return pxobj*
PUBLIC_OBJECT
pxo_string_new:
	;
	push ebp
	mov ebp, esp
	
	; return pxo_new( 3*sizeof(int) )
	push PXSTRING_SIZE
	call pxo_new
	
	mov dword[eax + PXSTRING_OFFSET_BUFFER  ], 0 ;buffer
	mov dword[eax + PXSTRING_OFFSET_BUFFSIZE], 0 ;bufflen
	mov dword[eax + PXSTRING_OFFSET_STRLEN  ], 0 ;strlen
	
	mov dword[eax - PXO_SIZE + PXO_OFFSET_DESTRUCTOR], pxo_string_dtor
	
	leave
	ret 0
	
;private void pxo_string_dtor(pxstring* this)
pxo_string_dtor:
	;
	push ebp
	mov ebp, esp
	
	;free(this + PXSTRING_OFFSET_BUFFER)	
	mov eax, dword[ebp+8]
	push dword[eax + PXSTRING_OFFSET_BUFFER]
	call pxl_rtl_free
	;
	leave
	ret 4 ; 1 argument

; public void pxo_string_dup(pxobj* other)
;  Duplicate a PXString. Used whenever referencing constants (since otherwise
;   assigning a string from a constant would render it unmodifyable)
;  Arguably should be used to implement operator= for string types - in general,
;   should operator= perform a value copy?
;
; @param pxobj* other
; @return pxobj* new
PUBLIC_OBJECT
pxo_string_dup:
	;
	push ebp
	mov ebp, esp
	push ebx
	push ecx
	
	; [ebp-8] := new PXString;
	call pxo_string_new
	push eax ; ebp-8
	
	; new->strlen  = other->strlen
	; new->bufflen = other->strlen
	mov ebx, dword[ebp+8] ;other
	mov ecx, dword[ebx + PXSTRING_OFFSET_STRLEN]
	mov dword[eax + PXSTRING_OFFSET_BUFFSIZE], ecx
	mov ecx, dword[ebx + PXSTRING_OFFSET_STRLEN]
	mov dword[eax + PXSTRING_OFFSET_STRLEN], ecx
	
	; new->buffer  = malloc(other->strlen)
	push dword[ebx + PXSTRING_OFFSET_STRLEN]
	call pxl_rtl_malloc
	
	push eax ;ebp-16
	mov eax, dword[ebp-12]
	mov ecx, dword[ebp-16]
	mov dword[eax + PXSTRING_OFFSET_BUFFER], ecx
		
	; memcpy(dest, source, numbytes)
	push dword[eax + PXSTRING_OFFSET_BUFFSIZE] ;numbytes
	push dword[ebx + PXSTRING_OFFSET_BUFFER]   ;source
	push dword[eax + PXSTRING_OFFSET_BUFFER]   ;dest
	call pxl_rtl_memcpy
	
	; return [dest]
	mov eax, dword[ebp-12]
	
	;
	pop ecx
	pop ebx
	leave
	ret 4 ; one argument
	
	
; public void pxo_string_append
;  If buffer size is insufficient and less than 8192, increase to next biggest power of two
;  If buffer size is insufficient and greater than 8192, add 8192. or something like that
; @param pxobj* self
; @param pxobj* other
; @return void
PUBLIC_OBJECT
pxo_string_append:
	;
	push ebp
	mov ebp, esp
	push ebx
	
	; ebx := next-greatest-power-of-two( eax );
	; http://stackoverflow.com/a/1322548
	xor ebx, ebx
	shr eax, 1 ; 1
	or ebx, eax
	shr eax, 1 ; 2
	or ebx, eax
	shr eax, 2 ; 4
	or ebx, eax
	shr eax, 4 ; 8
	or ebx, eax
	shr eax, 8 ; 16
	or ebx, eax
	add ebx, 1
	
	; 
	pop ebx
	leave
	ret 8 ; two arguments
	

; public void pxo_string_substr
; @param pxobj* self
; @param int start
; @param int count = -1
; @return pxobj*

; public void pxo_string_trace
; @param pxobj* self
; @return void
PUBLIC_OBJECT
pxo_string_trace:
	push ebp
	mov ebp, esp
	push ebx
	
	; if (! pxobj->buffer) return;
	mov ebx, dword[ebp+8] 
	cmp dword[ebx], 0
	je pxo_string_trace_complete
	
	; eax := GetStdHandle(STD_INPUT_HANDLE);
	push dword -11 ; 0xFFFFFFF5
	call _GetStdHandle@4
	
	; WriteFile(stdout, pxobj->buffer, pxobj->strlen, NULL, NULL);
	push 0
	push 0
	push dword[ebx + PXSTRING_OFFSET_STRLEN]
	push dword[ebx + PXSTRING_OFFSET_BUFFER]
	push eax ; stdout
	call _WriteFile@20
	
pxo_string_trace_complete:
	pop ebx
	leave
	ret 4 ; one argument
